﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.IO;

namespace SourceCounter
{
    public partial class FormMain : Form
    {
        public FormMain()
        {
            InitializeComponent();
        }

        private void btnOpenFolder_Click(object sender, EventArgs e)
        {
            if (folderBrowserDialog.ShowDialog() == DialogResult.OK)
            {
                this.Cursor = Cursors.WaitCursor;
                BuildFolderTree(folderBrowserDialog.SelectedPath, textExclude.Text, textFilter.Text);
                this.Cursor = Cursors.Default;
            }

            listViewFiles.Items.Clear();
            btnAnalyze.Enabled = treeViewFolders.Nodes.Count > 0;
            btnSaveFolder.Enabled = false;

            if (treeViewFolders.Nodes.Count > 0)
            {
                treeViewFolders.Nodes[0].Expand();
                treeViewFolders.SelectedNode = treeViewFolders.Nodes[0];
            }
        }

        private void btnAnalyze_Click(object sender, EventArgs e)
        {
            toolStripProgressBar.Value = 0;
            toolStripProgressBar.Maximum = treeViewFolders.SelectedNode.Nodes.Count;
            Application.DoEvents();

            this.Cursor = Cursors.WaitCursor;
            Analyze(treeViewFolders.SelectedNode);
            ShowFiles(treeViewFolders.SelectedNode);
            this.Cursor = Cursors.Default;
        }

        private void btnSaveFolders_Click(object sender, EventArgs e)
        {
            saveFileDialog.Filter = "Comma Separated Files (*.csv)|*.csv|All Files (*.*)|*.*";
            saveFileDialog.FileName = treeViewFolders.SelectedNode.Text + "_ByFolders.csv";

            if (saveFileDialog.ShowDialog() == DialogResult.OK)
            {
                this.Cursor = Cursors.WaitCursor;

                using (var outFile = new StreamWriter(saveFileDialog.FileName))
                {
                    outFile.WriteLine("Folder;Total Lines;Code Lines;File Size");
                    SaveFoldersData(treeViewFolders.SelectedNode, outFile);
                }

                this.Cursor = Cursors.Default;
            }
        }

        private void btnSaveFiles_Click(object sender, EventArgs e)
        {
            saveFileDialog.Filter = "Comma Separated Files (*.csv)|*.csv|All Files (*.*)|*.*";
            saveFileDialog.FileName = treeViewFolders.SelectedNode.Text + "_ByFiles.csv";

            if (saveFileDialog.ShowDialog() == DialogResult.OK)
            {
                toolStripProgressBar.Value = 0;
                toolStripProgressBar.Maximum = treeViewFolders.SelectedNode.Nodes.Count;
                Application.DoEvents();

                this.Cursor = Cursors.WaitCursor;

                using (var outFile = new StreamWriter(saveFileDialog.FileName))
                {
                    outFile.WriteLine("File;Total Lines;Code Lines;File Size");
                    foreach (TreeNode node in treeViewFolders.SelectedNode.Nodes)
                    {
                        SaveFilesData(node, outFile);
                        toolStripProgressBar.Increment(1);
                    }
                }

                this.Cursor = Cursors.Default;
            }
        }

        private void treeView_AfterSelect(object sender, TreeViewEventArgs e)
        {
            if (e.Node.Tag.GetType() == typeof(string))
                LoadFiles(e.Node, e.Node.Tag.ToString(), textFilter.Text);

            ShowFiles(e.Node);
        }

        private void treeViewFolders_Resize(object sender, EventArgs e)
        {
            toolStripProgressBar.Width = treeViewFolders.Width;
        }


        private void BuildFolderTree(string root, string exclude, string filter)
        {
            treeViewFolders.Nodes.Clear();
            DirectoryInfo dirInfo = new DirectoryInfo(root);

            toolStripProgressBar.Value = 0;
            toolStripProgressBar.Maximum = Directory.GetDirectories(root, "**", SearchOption.AllDirectories).Length;
            TreeNode tds = treeViewFolders.Nodes.Add(dirInfo.Name);
            tds.Tag = dirInfo.FullName;
            LoadSubDirectories(root, tds, exclude + ";", filter);
            toolStripProgressBar.Value = toolStripProgressBar.Maximum;
        }

        private int LoadSubDirectories(string dir, TreeNode td, string exclude, string filter)
        {
            int folderCount = 0;
            string[] subFolders = Directory.GetDirectories(dir);

            foreach (string folder in subFolders)
            {
                DirectoryInfo dirInfo = new DirectoryInfo(folder);
                if (!exclude.Contains(dirInfo.Name + ";"))
                {
                    if (CheckFiles(dirInfo.FullName, filter))
                    {
                        TreeNode tds = td.Nodes.Add(dirInfo.Name);
                        tds.Tag = dirInfo.FullName;
                        LoadSubDirectories(folder, tds, exclude, filter);
                        folderCount++;
                    }
                    else if (Directory.GetDirectories(dirInfo.FullName).Length > 0)
                    {
                        TreeNode tds = td.Nodes.Add(dirInfo.Name);
                        if (LoadSubDirectories(folder, tds, exclude, filter) == 0)
                        {
                            td.Nodes.Remove(tds);
                        }
                        else
                        {
                            tds.Tag = dirInfo.FullName;
                            folderCount++;
                        }
                    }
                }

                toolStripProgressBar.Increment(1);
            }

            return folderCount;
        }

        private bool CheckFiles(string dir, string filter)
        {
            string[] filters = filter.Split(';');

            foreach (string f in filters)
            {
                string[] Files = Directory.GetFiles(dir, f);

                if (Files.Length > 0)
                    return true;
            }

            return false;
        }

        private void LoadFiles(TreeNode node, string dir, string filter)
        {
            List<string> fileList = new List<string>();

            fileList.Add(dir); // First element is the FolderPath

            string[] filters = filter.Split(';');

            foreach (string f in filters)
            {
                string[] Files = Directory.GetFiles(dir, f);

                foreach (string file in Files)
                {
                    FileInfo fileInfo = new FileInfo(file);
                    fileList.Add(fileInfo.Name);
                }
            }

            node.Tag = fileList;
        }

        private void ShowFiles(TreeNode node)
        {
            statusLabelFolder.Text = node.Text;

            listViewFiles.Items.Clear();

            if (node.Tag.GetType() != typeof(List<string>))
                return;

            List<string> fileList = (List<string>)node.Tag;
            string[] folderStat = fileList[0].Split(';');

            if (folderStat.Count() == 4)
            {
                statusLabelLines.Text = "Lines: " + folderStat[1];
                statusLabelCode.Text = "Code: " + folderStat[2];
                statusLabelSize.Text = "Size: " + folderStat[3];
                btnSaveFolder.Enabled = true;
                btnSaveFiles.Enabled = true;
            }
            else
            {
                statusLabelLines.Text = "";
                statusLabelCode.Text = "";
                statusLabelSize.Text = "";
                btnSaveFolder.Enabled = false;
                btnSaveFiles.Enabled = false;
            }

            for (int f = 1; f < fileList.Count(); f++) // Skip FolderPath
            {
                string[] columns = fileList[f].Split(';');

                ListViewItem item = listViewFiles.Items.Add(columns[0]);
                if (columns.Count() == listViewFiles.Columns.Count)
                {
                    for (int c = 1; c < columns.Count(); c++)
                        item.SubItems.Add(columns[c]);
                }

                item.ImageIndex = 2;
            }
        }

        private void Analyze(TreeNode root)
        {
            int totalLinesCount = 0;
            int totalCodeCount = 0;
            int totalFolderSize = 0;

            AnalyzeFolder(root, 0, ref totalLinesCount, ref totalCodeCount, ref totalFolderSize);
        }

        private bool AnalyzeFolder(TreeNode root, int level, ref int totalLinesCount, ref int totalCodeCount, ref int totalFolderSize)
        {
            if (level == 1)
            {
                toolStripProgressBar.Increment(1);
                statusLabelFolder.Text = root.Text;
                Application.DoEvents();
            }

            if (root.Tag.GetType() == typeof(string))
                LoadFiles(root, root.Tag.ToString(), textFilter.Text);

            if (root.Tag.GetType() == typeof(List<string>))
            {
                List<string> fileList = (List<string>)root.Tag;
                
                string[] folderStat = fileList[0].Split(';');
                if (folderStat.Count() == 1) // Need Analysis
                {
                    int folderLinesCount = 0;
                    int folderCodeCount = 0;
                    int folderTotSize = 0;

                    string folderPath = folderStat[0];

                    for (int f = 1; f < fileList.Count(); f++) // Skip FolderPath
                    {
                        string[] columns = fileList[f].Split(';');

                        if (columns.Count() == 1) // Need Analysis
                        {
                            string fileName = Path.Combine(folderPath, columns[0]);

                            int linesCount = 0;
                            int codeCount = 0;
                            int totSize = 0;

                            try
                            {
                                using (var srcFile = new StreamReader(fileName, true))
                                {
                                    while (!srcFile.EndOfStream)
                                    {
                                        linesCount++;

                                        string line = srcFile.ReadLine();

                                        if (line != string.Empty)
                                            totSize += line.Length;

                                        line = line.TrimStart(new char[] { ' ', '\t' });
                                        if (line != string.Empty)
                                        {
                                            string ini = line.Length > 1 ? line.Substring(0, 2) : "  ";
                                            if (ini != "//" && ini != "/*" && ini != "*/")
                                                codeCount++;
                                        }
                                    }
                                }

                                fileList[f] = columns[0] + ";" +
                                              linesCount.ToString("N0") + ";" +
                                              codeCount.ToString("N0") + ";" +
                                              totSize.ToString("N0");

                                folderLinesCount += linesCount;
                                folderCodeCount += codeCount;
                                folderTotSize += totSize;
                            }
                            catch (Exception ex)
                            {
                                if (MessageBox.Show(ex.Message, "Error", MessageBoxButtons.AbortRetryIgnore, MessageBoxIcon.Error) == DialogResult.Abort)
                                    return false;
                            }
                        }
                    }

                    foreach (TreeNode node in root.Nodes)
                        if (!AnalyzeFolder(node, level + 1, ref folderLinesCount, ref folderCodeCount, ref folderTotSize))
                            return false;

                    fileList[0] = folderPath + ";" +
                                  folderLinesCount.ToString("N0") + ";" +
                                  folderCodeCount.ToString("N0") + ";" +
                                  folderTotSize.ToString("N0");

                    totalLinesCount += folderLinesCount;
                    totalCodeCount += folderCodeCount;
                    totalFolderSize += folderTotSize;
                }
            }

            return true;
        }

        private void SaveFoldersData(TreeNode root, StreamWriter stream)
        {
            if (root.Tag.GetType() == typeof(List<string>))
            {
                List<string> fileList = (List<string>)root.Tag;
                stream.WriteLine(fileList[0]);
            }

            foreach (TreeNode node in root.Nodes)
                SaveFoldersData(node, stream);
        }

        private void SaveFilesData(TreeNode root, StreamWriter stream)
        {
            if (root.Tag.GetType() == typeof(List<string>))
            {
                List<string> fileList = (List<string>)root.Tag;
                string[] folderStat = fileList[0].Split(';');
                string folderPath = folderStat[0];

                for (int f = 1; f < fileList.Count(); f++) // Skip FolderPath
                {
                    string line = Path.Combine(folderPath, fileList[f]);
                    stream.WriteLine(line);
                }
            }

            foreach (TreeNode node in root.Nodes)
                SaveFilesData(node, stream);
        }

    }
}
